#include "dhcpd.h"
#include <unistd.h>
#include <time.h>
#include <pwd.h>
#define group real_group
#include <grp.h>
#undef group

#include <sys/capability.h>
#include <sys/prctl.h>

static int minimized = 0;

void
dhcpd_priv_minimize(const char *user, const char *dir)
{
	struct passwd *pw;
	uid_t   uid;
	gid_t   gid;

	if (!*user)
		return;

	if (!(pw = getpwnam(user)))
		log_fatal("Failed to lower privileges: getpwnam: %s", user);

	uid = pw->pw_uid;
	gid = pw->pw_gid;

	tzset();

	if (initgroups(user, gid))
		log_fatal("Failed to lower privileges: initgroups: %s/%u: %m",
			  user, (unsigned) gid);

	endpwent();

	if (*dir && (chdir(dir) || chroot(".")))
		log_fatal("Failed to lower privileges: chroot: %s: %m", dir);

	if (setgid(gid))
		log_fatal("Failed to lower privileges: setgid: %s/%u: %m",
			  user, (unsigned) gid);

	if (prctl(PR_SET_KEEPCAPS, 1))
		log_fatal("Failed to lower privileges: prctl: %m");

	if (setreuid(uid, uid))
		log_fatal("Failed to lower privileges: setreuid: %s/%u: %m",
			  user, (unsigned) uid);

	cap_t   caps = cap_from_text("cap_net_bind_service=ep");

	if (!caps)
		log_fatal("Failed to lower privileges: cap_from_text: %m");

	if (cap_set_proc(caps) < 0)
		log_fatal("Failed to lower privileges: cap_set_proc: %m");

	cap_free(caps);
	minimized = 1;
}

static void
drop_minimized(void)
{
	cap_t   caps = cap_from_text("all-ep");
	if (!caps)
		log_fatal("Failed to lower privileges: cap_from_text: %m");

	if (cap_set_proc(caps) < 0)
		log_fatal("Failed to lower privileges: cap_set_proc: %m");

	cap_free(caps);
}

static void
drop_all(const char *user, const char *dir)
{
	struct passwd *pw;
	uid_t   uid;
	gid_t   gid;

	if (!(pw = getpwnam(user)))
		log_fatal("Failed to lower privileges: getpwnam: %s", user);

	uid = pw->pw_uid;
	gid = pw->pw_gid;

	tzset();

	if (initgroups(user, gid))
		log_fatal("Failed to lower privileges: initgroups: %s/%u: %m",
			  user, (unsigned) gid);

	endpwent();

	if (*dir && (chdir(dir) || chroot(".")))
		log_fatal("Failed to lower privileges: chroot: %s: %m", dir);

	if (setgid(gid))
		log_fatal("Failed to lower privileges: setgid: %s/%u: %m",
			  user, (unsigned) gid);

	if (setuid(uid))
		log_fatal("Failed to lower privileges: setuid: %s/%u: %m",
			  user, (unsigned) uid);
}


void
dhcpd_priv_drop(const char *user, const char *dir)
{
	if (!*user)
		return;

	if (minimized)
		drop_minimized();
	else
		drop_all(user, dir);
}
